from pydantic import BaseConfig
from pydantic.dataclasses import dataclass
from datetime import datetime
from sqlalchemy import func
from sqlalchemy.orm.query import Query
from typing import List, Optional

from rfhub2.db.base import KeywordStatistics
from rfhub2.db.repository.base_repository import BaseRepository
from rfhub2.db.repository.ordering import OrderingItem


class Config(BaseConfig):
    orm_mode = True


@dataclass(config=Config)
class AggregatedKeywordStatistics:

    times_used: int
    total_elapsed: int
    avg_elapsed: float
    min_elapsed: int
    max_elapsed: int


@dataclass
class KeywordStatisticsFilterParams:
    collection: str
    keyword: Optional[str] = None
    execution_time: Optional[datetime] = None
    execution_time_from: Optional[datetime] = None
    execution_time_to: Optional[datetime] = None


class KeywordStatisticsRepository(BaseRepository):
    @property
    def _items(self) -> Query:
        return self.session.query(KeywordStatistics)

    @staticmethod
    def filter_criteria(params: KeywordStatisticsFilterParams):
        filter_criteria = []
        if params.collection:
            filter_criteria.append(KeywordStatistics.collection == params.collection)
        if params.keyword:
            filter_criteria.append(KeywordStatistics.keyword == params.keyword)
        if params.execution_time:
            filter_criteria.append(
                KeywordStatistics.execution_time == params.execution_time
            )
        else:
            if params.execution_time_from:
                filter_criteria.append(
                    KeywordStatistics.execution_time >= params.execution_time_from
                )
            if params.execution_time_to:
                filter_criteria.append(
                    KeywordStatistics.execution_time <= params.execution_time_to
                )
        return filter_criteria

    def get_aggregated(
        self, filter_params: KeywordStatisticsFilterParams
    ) -> AggregatedKeywordStatistics:
        result = (
            self.session.query(
                func.coalesce(func.sum(KeywordStatistics.times_used), 0).label(
                    "times_used"
                ),
                func.coalesce(func.sum(KeywordStatistics.total_elapsed), 0).label(
                    "total_elapsed"
                ),
                func.coalesce(
                    func.sum(KeywordStatistics.total_elapsed)
                    / func.sum(KeywordStatistics.times_used),
                    0,
                ).label("avg_elapsed"),
                func.coalesce(func.min(KeywordStatistics.min_elapsed), 0).label(
                    "min_elapsed"
                ),
                func.coalesce(func.max(KeywordStatistics.max_elapsed), 0).label(
                    "max_elapsed"
                ),
            )
            .filter(*self.filter_criteria(filter_params))
            .first()
        )
        return AggregatedKeywordStatistics(*result)

    def get_many(
        self,
        *,
        filter_params: KeywordStatisticsFilterParams,
        skip: int = 0,
        limit: int = 100,
        ordering: List[OrderingItem] = None,
    ) -> List[KeywordStatistics]:
        return [
            stat.to_model()
            for stat in (
                self._items.filter(*self.filter_criteria(filter_params))
                .order_by(*KeywordStatistics.ordering_criteria(ordering))
                .offset(skip)
                .limit(limit)
                .all()
            )
        ]

    def add_many(self, items: List[KeywordStatistics]) -> int:
        count: int = len(items)
        self.session.bulk_save_objects(items)
        self.session.commit()
        return count

    def delete_many(
        self, filter_params: Optional[KeywordStatisticsFilterParams] = None
    ) -> int:
        filter_criteria = self.filter_criteria(filter_params) if filter_params else []
        row_count = self._items.filter(*filter_criteria).delete()
        self.session.commit()
        return row_count



from pathlib import Path
from typing import Dict, List, Set, Tuple, Union

from rfhub2.cli.api_client import Client
from rfhub2.cli.keywords.keywords_extractor import (
    CollectionUpdateWithKeywords,
    KeywordsExtractor,
)
from rfhub2.model import (
    Collection,
    CollectionUpdate,
    KeywordCreate,
    KeywordUpdate,
    NestedKeyword,
)


class KeywordsImporter:
    def __init__(
        self,
        client: Client,
        paths: Tuple[Union[Path, str], ...],
        no_installed_keywords: bool,
        load_mode: str,
    ) -> None:
        self.client = client
        self.paths = paths
        self.no_installed_keywords = no_installed_keywords
        self.load_mode = load_mode

    def delete_all_collections(self) -> Set[int]:
        """
        Deletes all existing collections.
        """
        collections_id = set()
        while len(self.client.get_collections()) > 0:
            collections_id.update(self._delete_collections())
        return collections_id

    def get_all_collections(self) -> List[Collection]:
        """Gets all collections from application"""
        collections = self.client.get_collections(0, 999999)
        return self._convert_json_to_collection(collections)

    def _delete_collections(self) -> Set[int]:
        """
        Helper method to delete all existing callections.
        """
        collections = self.client.get_collections()
        collections_id = {collection["id"] for collection in collections}
        for id in collections_id:
            self.client.delete_collection(id)
        return collections_id

    def import_data(self) -> Tuple[int, int]:
        """
        Wrapper for import_libraries and import_statistics to unify modules.
        :return: Number of libraries and keyword loaded
        """
        return self.import_libraries()

    def import_libraries(self) -> Tuple[int, int]:
        """
        Import libraries to application from paths specified when invoking client.
        :return: Number of libraries and keywords loaded
        """
        keywords_extractor = KeywordsExtractor(self.paths, self.no_installed_keywords)
        libraries_paths = keywords_extractor.get_libraries_paths()
        collections = keywords_extractor.create_collections(libraries_paths)
        if self.load_mode == "append":
            loaded_collections = self.add_collections(collections)
        elif self.load_mode == "insert":
            self.delete_all_collections()
            loaded_collections = self.add_collections(collections)
        else:
            existing_collections = self.get_all_collections()
            loaded_collections = self.update_collections(
                existing_collections, collections
            )
            if self.load_mode == "merge":
                self.delete_outdated_collections(
                    existing_collections, collections, remove_not_matched=False
                )
            else:
                self.delete_outdated_collections(existing_collections, collections)
        return len(loaded_collections), sum(d["keywords"] for d in loaded_collections)

    def update_collections(
        self,
        existing_collections: List[Collection],
        new_collections: List[CollectionUpdateWithKeywords],
    ) -> List[Dict[str, int]]:
        """
        Updates collections already existing in app.
        :param existing_collections: List of existing collections object
        :param new_collections: List of new collections object
        :return: list of dictionaries with collection name and number of keywords.
        """
        collections_to_update = self._get_collections_to_update(
            existing_collections, new_collections
        )
        collections_to_insert = self._get_new_collections(
            existing_collections, new_collections
        )
        return self.add_collections(collections_to_update + collections_to_insert)

    def delete_outdated_collections(
        self,
        existing_collections: List[Collection],
        new_collections: List[CollectionUpdateWithKeywords],
        remove_not_matched: bool = True,
    ) -> Set[int]:
        """Deletes outdated collections
        :param existing_collections: List of existing collections
        :param new_collections: List of new collections
        :param remove_not_matched: removes not_matched collection paths as well
        :return Set of deleted collections_id
        """
        collections_to_delete = self._get_outdated_collections_ids(
            existing_collections, new_collections
        )
        if remove_not_matched:
            collections_to_delete |= self._get_obsolete_collections_ids(
                existing_collections, new_collections
            )
        for collection in collections_to_delete:
            self.client.delete_collection(collection)
        return collections_to_delete

    def add_collections(
        self, collections: List[CollectionUpdateWithKeywords]
    ) -> List[Dict[str, int]]:
        """
        Adds collections and keywords from provided list to app.
        :param collections: List of collections object
        :return: list of dictionaries with collection name and number of keywords.
        """
        loaded_collections = []
        for collection in collections:
            coll_req = self.client.add_collection(collection.collection)
            if coll_req[0] != 201:
                print(coll_req[1]["detail"])
                raise StopIteration
            collection_id = coll_req[1]["id"]
            for keyword in collection.keywords:
                keyword = KeywordCreate(
                    **{**keyword.dict(), "collection_id": collection_id}
                )
                self.client.add_keyword(keyword)
            loaded_collections.append(
                {
                    "name": collection.collection.name,
                    "keywords": len(collection.keywords),
                }
            )
            print(
                f"{collection.collection.name} library with {len(collection.keywords)} keywords loaded."
            )
        return loaded_collections

    @staticmethod
    def _get_obsolete_collections_ids(
        existing_collections: List[Collection],
        new_collections: List[CollectionUpdateWithKeywords],
    ) -> Set[int]:
        """Returns set of collection ids that were found in application but not in paths"""
        new_collections_paths = {
            new_collection.collection.path for new_collection in new_collections
        }
        return {
            existing_collection.id
            for existing_collection in existing_collections
            if existing_collection.path not in new_collections_paths
        }

    @staticmethod
    def _get_outdated_collections_ids(
        existing_collections: List[Collection],
        new_collections: List[CollectionUpdateWithKeywords],
    ) -> Set[int]:
        """Returns set of collection ids that were found in application but are outdated"""
        outdated_collections = set()
        if len(existing_collections) > 0:
            for new_collection in new_collections:
                for existing_collection in existing_collections:
                    if KeywordsImporter._collection_path_and_name_match(
                        new_collection.collection, existing_collection
                    ) and KeywordsImporter._library_or_resource_changed(
                        new_collection, existing_collection
                    ):
                        outdated_collections.add(existing_collection.id)
        return outdated_collections

    @staticmethod
    def _get_collections_to_update(
        existing_collections: List[Collection],
        new_collections: List[CollectionUpdateWithKeywords],
    ) -> List[CollectionUpdateWithKeywords]:
        """Returns list of collections to update that were found in paths and application"""
        collections_to_update = []
        if len(existing_collections) >= 0:
            for new_collection in new_collections:
                for existing_collection in existing_collections:
                    if KeywordsImporter._collection_path_and_name_match(
                        new_collection.collection, existing_collection
                    ) and KeywordsImporter._library_or_resource_changed(
                        new_collection, existing_collection
                    ):
                        collections_to_update.append(new_collection)
        return collections_to_update

    @staticmethod
    def _get_new_collections(
        existing_collections: List[Collection],
        new_collections: List[CollectionUpdateWithKeywords],
    ) -> List[CollectionUpdateWithKeywords]:
        """Returns list of collections to insert that were found in paths but not in application"""
        existing_collections_paths = {
            existing_collection.path for existing_collection in existing_collections
        }
        return [
            new_collection
            for new_collection in new_collections
            if new_collection.collection.path not in existing_collections_paths
        ]

    @staticmethod
    def _collection_path_and_name_match(
        new_collection: CollectionUpdate, existing_collection: Collection
    ) -> bool:
        return (
            new_collection.name == existing_collection.name
            and new_collection.path == existing_collection.path
        )

    @staticmethod
    def _library_or_resource_changed(
        new_collection: CollectionUpdateWithKeywords, existing_collection: Collection
    ) -> bool:
        if new_collection.collection.type.lower() == "library":
            return new_collection.collection.version != existing_collection.version
        else:
            return (
                any(
                    keyword not in new_collection.keywords
                    for keyword in KeywordsImporter._convert_keywords_to_keywords_update(
                        existing_collection.keywords
                    )
                )
                or any(
                    keyword
                    not in KeywordsImporter._convert_keywords_to_keywords_update(
                        existing_collection.keywords
                    )
                    for keyword in new_collection.keywords
                )
                or KeywordsImporter._library_or_resource_doc_changed(
                    new_collection.collection, existing_collection
                )
            )

    @staticmethod
    def _convert_keywords_to_keywords_update(
        keywords: List[NestedKeyword]
    ) -> List[KeywordUpdate]:
        """Convert list of Keywords object to List of KeywordUpdate object"""
        return [
            KeywordUpdate(
                name=keyword.name, doc=keyword.doc, args=keyword.args, tags=keyword.tags
            )
            for keyword in keywords
        ]

    @staticmethod
    def _convert_json_to_collection(collections: Dict) -> List[Collection]:
        for collection in collections:
            collection["keywords"] = [
                NestedKeyword(**keyword) for keyword in collection["keywords"]
            ]
        return [Collection(**collection) for collection in collections]

    @staticmethod
    def _library_or_resource_doc_changed(
        new_collection: CollectionUpdate, existing_collection: Collection
    ) -> bool:
        """Returns true if collection overall documentation has changed.
        Does not check for keywords changes"""
        reduced_existing_collection = CollectionUpdate(
            name=existing_collection.name,
            type=existing_collection.type,
            version=existing_collection.version,
            scope=existing_collection.scope,
            args=existing_collection.named_args,
            path=existing_collection.path,
            doc=existing_collection.doc,
            doc_format=existing_collection.doc_format,
        )
        return new_collection != reduced_existing_collection



from collections import Counter
from dataclasses import dataclass
from datetime import datetime
from itertools import groupby
from typing import Callable, Dict, Iterable, List, Tuple
from xml.etree.ElementTree import Element, parse

from rfhub2.model import KeywordStatistics


@dataclass
class XmlKeyword:
    library: str
    name: str
    elapsed: int


@dataclass
class ElapsedStats:
    total_elapsed: int
    min_elapsed: int
    max_elapsed: int


StatsKey = Tuple[str, str]


def stats_key(keyword: XmlKeyword) -> StatsKey:
    return keyword.library, keyword.name


class StatisticsExtractor:
    def __init__(self, path: str):
        self.path: str = path
        self.source_time_format: str = "%Y%m%d %H:%M:%S.%f"
        self.destination_time_format: str = "%Y-%m-%d %H:%M:%S.%f"

    def compute_statistics(self) -> List[KeywordStatistics]:
        """
        Returns list of KeywordStatistics extracted from single output.xml file
        """
        keywords = self.parse_xml_keywords()
        return self.aggregate_statistics(keywords)

    def aggregate_statistics(
        self, keywords: List[XmlKeyword]
    ) -> List[KeywordStatistics]:
        """
        Returns list of KeywordStatistics grouped by library/keyword pair.
        """
        execution_time = self.get_execution_time()
        elapsed_stats = self.get_elapsed_stats(keywords)
        statistics = []
        for key, count in Counter(stats_key(keyword) for keyword in keywords).items():
            stats = elapsed_stats[key]
            statistics.append(
                KeywordStatistics(
                    collection=key[0],
                    keyword=key[1],
                    execution_time=execution_time,
                    times_used=count,
                    total_elapsed=stats.total_elapsed,
                    min_elapsed=stats.min_elapsed,
                    max_elapsed=stats.max_elapsed,
                )
            )
        return statistics

    @staticmethod
    def get_elapsed_stats(keywords: List[XmlKeyword]) -> Dict[StatsKey, ElapsedStats]:
        """
        Returns dict of keyword execution stats such as min, max and total elapsed time
        grouped by library/keyword pair.
        """
        result = {}
        for key, group in groupby(sorted(keywords, key=stats_key), stats_key):
            elapsed_times = [kw.elapsed for kw in group]
            result[key] = ElapsedStats(
                total_elapsed=sum(elapsed_times),
                min_elapsed=min(elapsed_times),
                max_elapsed=max(elapsed_times),
            )
        return result

    def parse_xml_keywords(self) -> List[XmlKeyword]:
        """
        Returns list of XmlKeyword objects containing keywords execution data.
        """
        xml_keywords = parse(self.path).findall(".//kw")
        return [
            XmlKeyword(
                xml_keyword.attrib.get("library"),
                xml_keyword.attrib.get("name"),
                self.calc_elapsed(xml_keyword),
            )
            for xml_keyword in xml_keywords
            if xml_keyword.attrib.get("library") is not None
        ]

    def datetime_from_attribute_agg(
        self, element: Element, attr: str, agg_func: Callable[[Iterable[str]], str]
    ) -> datetime:
        return datetime.strptime(
            agg_func(
                tag.attrib.get(attr) for tag in element.iter() if tag.tag == "status"
            ),
            self.source_time_format,
        )

    def calc_elapsed(self, element: Element) -> int:
        return int(
            1000
            * (
                self.datetime_from_attribute_agg(element, "endtime", max)
                - self.datetime_from_attribute_agg(element, "starttime", min)
            ).total_seconds()
        )

    def get_execution_time(self) -> str:
        """
        Returns execution time extracted form robot output.xml file.
        """
        return datetime.strftime(
            datetime.strptime(
                parse(self.path).getroot().attrib.get("generated"),
                self.source_time_format,
            ),
            self.destination_time_format,
        )



from typing import Optional, Tuple

IN_TOKEN = " in:"
NAME_TOKEN = "name:"
TAG_TOKEN = "tags:"
WILDCARD = "*"


class SearchParams:

    DEFAULT = (WILDCARD, None, True, False)

    def __init__(self, pattern: str = WILDCARD) -> None:
        if not pattern or pattern == WILDCARD:
            result = self.DEFAULT
        else:
            result = self.extract_params(pattern)
        self.raw_pattern = pattern
        self.pattern, self.collection_name, self.use_doc, self.use_tags = result

    @staticmethod
    def extract_params(raw_pattern: str) -> Tuple[str, Optional[str], bool, bool]:
        pattern, collection_name, use_doc, use_tags = (
            raw_pattern.strip().lower(),
            None,
            True,
            False,
        )
        if pattern.startswith(NAME_TOKEN):
            pattern = pattern[5:].strip()
            use_doc = False
        query, sep, col_name = pattern.partition(IN_TOKEN)
        if sep == IN_TOKEN and col_name:
            pattern = query.strip()
            collection_name = col_name.strip()
        query, sep, tag_name = pattern.partition(TAG_TOKEN)
        if sep == TAG_TOKEN:
            pattern = tag_name.strip()
            use_doc, use_tags = False, True
        return pattern, collection_name, use_doc, use_tags



"""Initial model

Revision ID: c54a916ec6c8
Revises: 
Create Date: 2020-04-01 21:02:20.117380

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "c54a916ec6c8"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "collection",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("type", sa.Text(), nullable=True),
        sa.Column("version", sa.Text(), nullable=True),
        sa.Column("scope", sa.Text(), nullable=True),
        sa.Column("named_args", sa.Text(), nullable=True),
        sa.Column("path", sa.Text(), nullable=True),
        sa.Column("doc", sa.Text(), nullable=True),
        sa.Column("doc_format", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_collection_name"), "collection", ["name"], unique=False)
    op.create_table(
        "keywordstatistics",
        sa.Column("collection", sa.Text(), nullable=False),
        sa.Column("keyword", sa.Text(), nullable=False),
        sa.Column("execution_time", sa.DateTime(timezone=True), nullable=False),
        sa.Column("times_used", sa.Integer(), nullable=True),
        sa.Column("total_elapsed", sa.Integer(), nullable=True),
        sa.Column("min_elapsed", sa.Integer(), nullable=True),
        sa.Column("max_elapsed", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("collection", "keyword", "execution_time"),
    )
    op.create_table(
        "keyword",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=True),
        sa.Column("doc", sa.Text(), nullable=True),
        sa.Column("args", sa.Text(), nullable=True),
        sa.Column("collection_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["collection_id"], ["collection.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_keyword_name"), "keyword", ["name"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_keyword_name"), table_name="keyword")
    op.drop_table("keyword")
    op.drop_table("keywordstatistics")
    op.drop_index(op.f("ix_collection_name"), table_name="collection")
    op.drop_table("collection")
    # ### end Alembic commands ###



from fastapi import APIRouter

from rfhub2.api.endpoints import (
    collections,
    healthcheck,
    keywords,
    keyword_statistics,
    version,
)

api_router = APIRouter()
api_router.include_router(healthcheck.router, prefix="/health", tags=["healthcheck"])
api_router.include_router(
    collections.router, prefix="/collections", tags=["collections"]
)
api_router.include_router(keywords.router, prefix="/keywords", tags=["keywords"])
api_router.include_router(
    keyword_statistics.router, prefix="/statistics/keywords", tags=["statistics"]
)
api_router.include_router(version.router, prefix="/version", tags=["version"])



from rfhub2.db.model.base_class import Base
from rfhub2.db.model.keyword import Keyword
from rfhub2.db.model.collection import Collection
from rfhub2.db.model.keyword_statistics import KeywordStatistics
from rfhub2.db.model.suite import Suite
from rfhub2.db.model.suite_rel import SuiteRel
from rfhub2.db.model.test_case import TestCase



from alembic.config import Config
from alembic.runtime.migration import MigrationContext
from alembic import command
from pathlib import Path
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.engine.reflection import Inspector


def has_tables(connection: Connection) -> bool:
    inspector = Inspector(connection)
    tables = inspector.get_table_names()
    return len(tables) > 0


def has_revision(connection: Connection) -> bool:
    context = MigrationContext.configure(connection)
    current_rev = context.get_current_revision()
    return current_rev is not None


def migrate_db(engine: Engine) -> None:
    config_path = Path(__file__).resolve().parent.parent / "alembic.ini"
    alembic_cfg = Config(config_path)
    with engine.begin() as connection:
        # check if database has not been migrated yet with alembic
        if has_tables(connection) and not has_revision(connection):
            # stamp database with the initial revision id, representing pre-alembic database schema
            command.stamp(alembic_cfg, "c54a916ec6c8")
        command.upgrade(alembic_cfg, "head")



from itertools import zip_longest
from typing import Dict, List, Optional, Tuple

from sqlalchemy import or_, Column
from sqlalchemy.orm.query import Query
from sqlalchemy.sql import func
from sqlalchemy.orm.session import Session

from rfhub2.db.base import Suite, SuiteRel, TestCase
from rfhub2.model import (
    KeywordRefList,
    Suite as ModelSuite,
    SuiteHierarchy,
    SuiteHierarchyWithId,
    SuiteMetadata,
)
from rfhub2.db.repository.base_repository import BaseRepository
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2.db.repository.query_utils import glob_to_sql


class SuiteRepository(BaseRepository):
    def __init__(self, db_session: Session):
        super().__init__(db_session)
        self.test_counts = (
            self.session.query(
                SuiteRel.parent_id.label("suite_id"),
                func.count(TestCase.id).label("test_count"),
            )
            .outerjoin(SuiteRel, TestCase.suite_id == SuiteRel.child_id)
            .group_by(SuiteRel.parent_id)
            .subquery()
        )

    @property
    def _items(self) -> Query:
        return self.items(parent_id=None)

    @property
    def custom_column_mapping(self) -> Dict[str, Column]:
        return {"test_count": self.test_count_column}

    @property
    def test_count_column(self) -> Column:
        return func.coalesce(self.test_counts.c.test_count, 0)

    @staticmethod
    def filter_criteria(
        pattern: Optional[str], root_only: bool, use_doc: bool, use_tags: bool
    ):
        filter_criteria = []
        if pattern:
            if use_tags:
                filter_criteria.append(Suite.tags.ilike(glob_to_sql(pattern)))
            else:
                filter_criteria.append(Suite.name.ilike(glob_to_sql(pattern)))
            if use_doc:
                filter_criteria = [
                    or_(filter_criteria[0], Suite.doc.ilike(glob_to_sql(pattern)))
                ]
        if root_only:
            filter_criteria.append(Suite.is_root.is_(True))
        return filter_criteria

    @staticmethod
    def from_row(row: Tuple[Suite, Optional[int], int]) -> ModelSuite:
        suite = row[0]
        return ModelSuite(
            id=suite.id,
            name=suite.name,
            longname=suite.longname,
            doc=suite.doc,
            source=suite.source,
            is_root=suite.is_root,
            rpa=suite.rpa,
            parent_id=row[1],
            test_count=row[2],
            keywords=KeywordRefList.parse_raw(suite.keywords),
            metadata=SuiteMetadata.parse_raw(suite.metadata_items),
        )

    def add_hierarchy(
        self, hierarchy: SuiteHierarchy
    ) -> Optional[SuiteHierarchyWithId]:
        try:
            hierarchies = self._add_suites([hierarchy], is_root=True)
            self._add_suite_rels(hierarchies, [])
            self.session.commit()
            return hierarchies[0]
        except Exception as e:
            self.session.rollback()
            raise e

    def delete_hierarchy(self, suite_id: int) -> int:
        suite_ids = (
            self.session.query(SuiteRel.child_id)
            .filter(SuiteRel.parent_id == suite_id)
            .subquery()
        )
        deleted = (
            self.session.query(Suite)
            .filter(Suite.id.in_(suite_ids))
            .delete(synchronize_session=False)
        )
        self.session.commit()
        return deleted

    def get(self, item_id: int) -> Optional[ModelSuite]:
        result = self._items.filter(Suite.id == item_id).first()
        if result:
            return self.from_row(result)

    def get_all(
        self,
        *,
        pattern: Optional[str] = None,
        parent_id: Optional[int] = None,
        root_only: bool = False,
        use_doc: bool = True,
        use_tags: bool = False,
        skip: int = 0,
        limit: int = 100,
        ordering: List[OrderingItem] = None,
    ) -> List[ModelSuite]:
        return [
            self.from_row(row)
            for row in (
                self.items(parent_id)
                .filter(*self.filter_criteria(pattern, root_only, use_doc, use_tags))
                .order_by(
                    *Suite.ordering_criteria(ordering, self.custom_column_mapping)
                )
                .offset(skip)
                .limit(limit)
                .all()
            )
        ]

    def items(self, parent_id: Optional[int] = None):
        parents = (
            self.session.query(SuiteRel.parent_id, SuiteRel.child_id)
            .filter(SuiteRel.degree == 1)
            .subquery()
        )
        query = (
            self.session.query(Suite, parents.c.parent_id, self.test_count_column)
            .outerjoin(parents, Suite.id == parents.c.child_id)
            .outerjoin(self.test_counts, Suite.id == self.test_counts.c.suite_id)
        )
        if parent_id:
            query = query.filter(parents.c.parent_id == parent_id)
        return query

    def _add_suites(
        self,
        hierarchies: List[SuiteHierarchy],
        is_root: bool,
        parent: Optional[SuiteHierarchyWithId] = None,
    ) -> List[SuiteHierarchyWithId]:
        if hierarchies:
            suites = [
                Suite.create(hierarchy, is_root, parent) for hierarchy in hierarchies
            ]
            self.session.add_all(suites)
            self.session.flush()
            hierarchies_with_id = [suite.to_hierarchy() for suite in suites]
            subhierarchies = [
                self._add_suites(h.suites, is_root=False, parent=h_id)
                for h, h_id in zip_longest(hierarchies, hierarchies_with_id)
            ]
            return [
                h.with_suites(ss)
                for h, ss in zip_longest(hierarchies_with_id, subhierarchies)
            ]
        else:
            return []

    def _add_suite_rels(
        self, hierarchies: List[SuiteHierarchyWithId], parents: List[int]
    ) -> None:
        for hierarchy in hierarchies:
            new_parents = [hierarchy.id] + parents
            self.session.add_all(
                [
                    SuiteRel(parent_id=parent, child_id=hierarchy.id, degree=degree)
                    for degree, parent in enumerate(new_parents)
                ]
            )
            self._add_suite_rels(hierarchy.suites, new_parents)



from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.exc import IntegrityError
from starlette.responses import Response
from typing import List

from rfhub2.api.utils.auth import is_authenticated
from rfhub2.api.utils.db import get_keyword_statistics_repository
from rfhub2.api.utils.order import get_ordering
from rfhub2.db.base import KeywordStatistics as DBStatistics
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2.db.repository.keyword_statistics_repository import (
    AggregatedKeywordStatistics,
    KeywordStatisticsFilterParams,
    KeywordStatisticsRepository,
)
from rfhub2.model import KeywordStatistics, StatisticsDeleted, StatisticsInserted

router = APIRouter()


class DuplicatedStatisticsException(HTTPException):
    def __init__(self):
        super(DuplicatedStatisticsException, self).__init__(
            status_code=400,
            detail="Records already exist for provided collection, keyword and execution_time",
        )


@router.get("/aggregated/", response_model=AggregatedKeywordStatistics)
def get_aggregated(
    *,
    repository: KeywordStatisticsRepository = Depends(
        get_keyword_statistics_repository
    ),
    filter_params: KeywordStatisticsFilterParams = Depends(),
):
    return repository.get_aggregated(filter_params)


@router.get("/", response_model=List[KeywordStatistics])
def get_statistics(
    *,
    repository: KeywordStatisticsRepository = Depends(
        get_keyword_statistics_repository
    ),
    filter_params: KeywordStatisticsFilterParams = Depends(),
    skip: int = 0,
    limit: int = 100,
    ordering: List[OrderingItem] = Depends(get_ordering),
):
    return repository.get_many(
        filter_params=filter_params, skip=skip, limit=limit, ordering=ordering
    )


@router.post("/", status_code=201)
def create_statistics(
    *,
    _: bool = Depends(is_authenticated),
    repository: KeywordStatisticsRepository = Depends(
        get_keyword_statistics_repository
    ),
    statistics: List[KeywordStatistics],
):
    db_statistics: List[DBStatistics] = [
        DBStatistics.create(stat) for stat in statistics
    ]
    try:
        inserted: int = repository.add_many(db_statistics)
        return StatisticsInserted(inserted=inserted)
    except IntegrityError:
        raise DuplicatedStatisticsException()


@router.delete("/")
def delete_statistics(
    *,
    response: Response,
    _: bool = Depends(is_authenticated),
    repository: KeywordStatisticsRepository = Depends(
        get_keyword_statistics_repository
    ),
    filter_params: KeywordStatisticsFilterParams = Depends(),
):
    deleted: int = repository.delete_many(filter_params)
    if deleted:
        response.status_code = 204
        return StatisticsDeleted(deleted=deleted)
    else:
        raise HTTPException(status_code=404)


@router.delete("/all/", status_code=204)
def delete_all_statistics(
    *,
    _: bool = Depends(is_authenticated),
    repository: KeywordStatisticsRepository = Depends(
        get_keyword_statistics_repository
    ),
):
    deleted: int = repository.delete_many()
    return StatisticsDeleted(deleted=deleted)



from collections import OrderedDict
from sqlalchemy import Column
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from typing import Dict, List

from rfhub2.db.repository.ordering import OrderingItem


class CustomBase(object):
    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    @classmethod
    def column_mapping(cls):
        return {col.name: col for col in cls.columns()}

    @classmethod
    def columns(cls):
        return cls.__table__._columns

    @classmethod
    def default_ordering(cls) -> List[OrderingItem]:
        return []

    @classmethod
    def ordering_criteria(
        cls, items: List[OrderingItem], custom_columns: Dict[str, Column] = dict()
    ):
        def criterion(col: Column, ordering_item: OrderingItem) -> Column:
            return col if ordering_item.asc else col.desc()

        if not items:
            items = cls.default_ordering()
        mapping = {**cls.column_mapping(), **custom_columns}
        return OrderedDict(
            (item.field, criterion(mapping[item.field], item))
            for item in items
            if item.field in mapping
        ).values()


Base = declarative_base(cls=CustomBase)



from typing import List, Optional, Tuple

from sqlalchemy import func
from sqlalchemy.orm import selectinload
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.elements import BinaryExpression

from rfhub2.db.base import Collection, KeywordStatistics
from rfhub2.model import CollectionWithStats, Collection as ModelCollection
from rfhub2.db.repository.base_repository import IdEntityRepository
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2.db.repository.query_utils import glob_to_sql


class CollectionRepository(IdEntityRepository):
    @property
    def _items(self) -> Query:
        return self.session.query(Collection).options(selectinload(Collection.keywords))

    @property
    def _items_with_stats(self) -> Query:
        collection_statistics = (
            self.session.query(
                (func.sum(KeywordStatistics.times_used)).label("times_used"),
                KeywordStatistics.collection,
            )
            .group_by(KeywordStatistics.collection)
            .subquery()
        )
        return (
            self.session.query(Collection, collection_statistics.c.times_used)
            .outerjoin(
                collection_statistics,
                Collection.name == collection_statistics.c.collection,
            )
            .options(selectinload(Collection.keywords))
        )

    def _id_filter(self, item_id: int) -> BinaryExpression:
        return Collection.id == item_id

    @staticmethod
    def filter_criteria(pattern: Optional[str], libtype: Optional[str]):
        filter_criteria = []
        if pattern:
            filter_criteria.append(Collection.name.ilike(glob_to_sql(pattern)))
        if libtype:
            filter_criteria.append(Collection.type.ilike(glob_to_sql(libtype)))
        return filter_criteria

    @staticmethod
    def from_stats_row(row: Tuple[Collection, int]) -> CollectionWithStats:
        collection = row[0]
        keywords = [kw.to_nested_model() for kw in collection.keywords]
        return CollectionWithStats(
            id=collection.id,
            name=collection.name,
            type=collection.type,
            version=collection.version,
            scope=collection.scope,
            named_args=collection.named_args,
            path=collection.path,
            doc=collection.doc,
            doc_format=collection.doc_format,
            html_doc=collection.html_doc,
            synopsis=collection.synopsis,
            keywords=keywords,
            times_used=row[1],
        )

    def get_all(
        self,
        *,
        pattern: Optional[str] = None,
        libtype: Optional[str] = None,
        skip: int = 0,
        limit: int = 100,
        ordering: List[OrderingItem] = None,
    ) -> List[ModelCollection]:
        return [
            collection.to_model()
            for collection in (
                self._items.filter(*self.filter_criteria(pattern, libtype))
                .order_by(*Collection.ordering_criteria(ordering))
                .offset(skip)
                .limit(limit)
                .all()
            )
        ]

    def get_all_with_stats(
        self,
        *,
        pattern: Optional[str] = None,
        libtype: Optional[str] = None,
        skip: int = 0,
        limit: int = 100,
        ordering: List[OrderingItem] = None,
    ) -> List[CollectionWithStats]:
        return [
            self.from_stats_row(row)
            for row in (
                self._items_with_stats.filter(*self.filter_criteria(pattern, libtype))
                .order_by(*Collection.ordering_criteria(ordering))
                .offset(skip)
                .limit(limit)
                .all()
            )
        ]

    def get_with_stats(self, item_id: int) -> Optional[CollectionWithStats]:
        result = self._items_with_stats.filter(self._id_filter(item_id)).first()
        if result:
            return self.from_stats_row(result)



from sqlalchemy import Boolean, Column, Index, Integer, Sequence, Text
from typing import List, Optional

from rfhub2.db.model.base_class import Base
from rfhub2.db.model.mixins import DocMixin
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2.model import (
    KeywordRefList,
    NestedSuite,
    SuiteHierarchy,
    SuiteHierarchyWithId,
    SuiteMetadata,
)


class Suite(Base, DocMixin):
    id = Column(Integer, Sequence("suite_id_seq"), primary_key=True)
    name = Column(Text, index=True)
    longname = Column(Text)
    doc = Column(Text)
    keywords = Column(Text)
    metadata_items = Column(Text)
    source = Column(Text)
    is_root = Column(Boolean)
    rpa = Column(Boolean)

    __table_args__ = (Index("ix_suite_longname", longname, unique=True),)

    def __str__(self):  # pragma: no cover
        return (
            f"Suite({self.id},{self.name},{self.longname},{self.doc},{self.source},"
            + f"{self.keywords},{self.metadata_items},{self.is_root},{self.rpa})"
        )

    def __repr__(self):  # pragma: no cover
        return str(self)

    @classmethod
    def default_ordering(cls) -> List[OrderingItem]:
        return [OrderingItem("longname")]

    @staticmethod
    def create(
        hierarchy: SuiteHierarchy,
        is_root: bool = True,
        parent: Optional[SuiteHierarchyWithId] = None,
    ) -> "Suite":
        return Suite(
            name=hierarchy.name,
            longname=f"{parent.longname}.{hierarchy.name}"
            if parent
            else hierarchy.name,
            doc=hierarchy.doc,
            source=hierarchy.source,
            keywords=hierarchy.keywords.json(),
            is_root=is_root,
            metadata_items=hierarchy.metadata.json(),
            rpa=hierarchy.rpa,
        )

    def to_hierarchy(self) -> SuiteHierarchyWithId:
        return SuiteHierarchyWithId(
            id=self.id,
            name=self.name,
            longname=self.longname,
            doc=self.doc,
            source=self.source,
            is_root=self.is_root,
            keywords=KeywordRefList.parse_raw(self.keywords),
            suites=[],
            metadata=SuiteMetadata.parse_raw(self.metadata_items),
            rpa=self.rpa,
        )

    def to_nested(self) -> NestedSuite:
        return NestedSuite(id=self.id, name=self.name, longname=self.longname)



from pathlib import Path
from progress.bar import IncrementalBar
from typing import List, Set, Tuple

from rfhub2.cli.api_client import Client
from .statistics_extractor import StatisticsExtractor
from rfhub2.model import KeywordStatistics, KeywordStatisticsList


class StatisticsImporter:
    def __init__(self, client: Client, paths: Tuple[Path, ...]) -> None:
        self.client = client
        self.paths = paths

    def import_data(self) -> Tuple[int, int]:
        """
        Wrapper for import_libraries and import_statistics to unify modules.
        :return: Number of libraries and keyword loaded
        """
        return self.import_statistics()

    def import_statistics(self) -> Tuple[int, int]:
        """
        Import keywords executions statistics such as min/max/total elapsed time,
        number of times used and execution timestamp.
        :return: Number of libraries and keyword loaded
        """
        execution_files = self.get_execution_files_paths()
        loaded_statistics = []
        progress_bar = IncrementalBar(
            "Sending statistics",
            max=len(execution_files),
            suffix="%(percent).1f%% - %(eta)ds, elapsed: %(elapsed)ds",
        )
        for execution_file in execution_files:
            statistics = StatisticsExtractor(execution_file).compute_statistics()
            loaded_statistics.append(self.add_statistics(statistics, execution_file))
            progress_bar.next()
        return (
            len([stat for stat in loaded_statistics if stat > 0]),
            sum(loaded_statistics),
        )

    def get_execution_files_paths(self) -> Set[Path]:
        """
        Traverses all given paths and returns set with paths
        pointing to RFWK output.xml files to import to app.
        :return: Set of Paths object pointing to output.xml files to import
        """
        return {
            p
            for path in self.paths
            for p in Path(path).rglob("*.xml")
            if self._is_valid_execution_file(p)
        }

    def add_statistics(
        self, statistics: List[KeywordStatistics], execution_file: Path
    ) -> int:
        """
        Adds statistics from provided list to app.
        :param statistics: List of KeywordStatistics objects
        :param execution_file: Path to file from statistics where extracted
        :return: number of statistics sent to app
        """
        stats_req = self.client.add_statistics(KeywordStatisticsList.of(statistics))
        if stats_req[0] == 201:
            return len(statistics)
        elif stats_req[0] != 400:
            print(stats_req[1]["detail"])
        else:
            print(f"""Records already exist for file from {execution_file}""")
        return 0

    @staticmethod
    def _is_valid_execution_file(path: Path) -> bool:
        """
        Checks if file is xml file containing apropriate string.
        This is simplified approach to quick check file.
        :param path:
        :return:
        """
        with open(path, "r", encoding="utf-8", errors="ignore") as f:
            # read the first few lines; if we don't see
            # what looks like robot generator tag, return false
            data = f.read(200)
            return "<robot generator=" in data.lower()



import click
from pathlib import Path
from typing import Tuple, Union

from rfhub2.cli.api_client import Client
from rfhub2.cli.keywords.keywords_importer import KeywordsImporter
from rfhub2.cli.statistics.statistics_importer import StatisticsImporter


@click.command()
@click.option(
    "-a",
    "--app-url",
    type=click.STRING,
    default="http://localhost:8000",
    help="Specifies IP, URI or host of rfhub2 web application. Default value is http://localhost:8000.",
)
@click.option(
    "-u",
    "--user",
    type=click.STRING,
    default="rfhub",
    help="Specifies rfhub2 user to authenticate on endpoints that requires that. Default value is rfhub.",
)
@click.option(
    "-p",
    "--password",
    type=click.STRING,
    default="rfhub",
    help="Specifies rfhub2 password to authenticate on endpoints that requires that. Default value is rfhub.",
)
@click.option(
    "--no-installed-keywords",
    type=click.BOOL,
    default=False,
    is_flag=True,
    help="Flag specifying if package should skip loading commonly installed libraries, "
    "such as such as BuiltIn, Collections, DateTime etc.",
)
@click.option(
    "--mode",
    "-m",
    type=click.Choice(["keywords", "statistics"], case_sensitive=False),
    default="keywords",
    help="""Choice parameter specifying what kind of data package should add:\n
             - `keywords` - default value, application is working with keywords documentation\n
             - `statistics` - application is working with data about keywords execution.""",
)
@click.option(
    "--load-mode",
    "-l",
    type=click.Choice(["insert", "append", "update", "merge"], case_sensitive=False),
    default="insert",
    help="""Choice parameter specifying in what load mode package should run:\n
             - `insert` - default value, removes all existing collections from app and add ones found in paths\n
             - `append` - adds collections found in paths without removal of existing ones\n
             - `update` - removes collections not found in paths, adds new ones and updates existing ones\n
             - `merge`  - adds new and updates only matched collections, does nothing with not matched ones.""",
)
@click.argument("paths", nargs=-1)
def main(
    app_url: str,
    user: str,
    password: str,
    paths: Tuple[Union[Path, str], ...],
    load_mode: str,
    mode: str,
    no_installed_keywords: bool,
) -> None:
    """Package to populate rfhub2 with robot framework keywords
       from libraries and resource files."""
    client = Client(app_url, user, password)
    if mode == "keywords":
        rfhub_importer = KeywordsImporter(
            client, paths, no_installed_keywords, load_mode
        )
        loaded_collections, loaded_keywords = rfhub_importer.import_data()
        print(
            f"\nSuccessfully loaded {loaded_collections} collections with {loaded_keywords} keywords."
        )
    elif mode == "statistics":
        rfhub_importer = StatisticsImporter(client, paths)
        loaded_files, loaded_statistics = rfhub_importer.import_data()
        print(
            f"\nSuccessfully loaded {loaded_files} files with {loaded_statistics} statistics."
        )



from fastapi import Query
from typing import List

from rfhub2.db.repository.ordering import OrderingItem


def get_ordering(order: List[str] = Query([])) -> List[OrderingItem]:
    def to_ordering_item(raw: str) -> OrderingItem:
        if raw.startswith("-"):
            return OrderingItem(raw[1:], False)
        else:
            return OrderingItem(raw)

    return [to_ordering_item(item) for item in order if item and item != "-"]



from typing import List, Optional, Tuple

from sqlalchemy import or_
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.elements import BinaryExpression

from rfhub2.db.base import Suite, TestCase
from rfhub2.model import KeywordRefList, TagList, TestCase as ModelTestCase
from rfhub2.db.repository.base_repository import IdEntityRepository
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2.db.repository.query_utils import glob_to_sql


class TestCaseRepository(IdEntityRepository):
    @property
    def _get_items(self) -> Query:
        return self.session.query(TestCase, Suite).filter(TestCase.suite_id == Suite.id)

    @property
    def _items(self) -> Query:
        return self.session.query(TestCase)

    def _id_filter(self, item_id: int) -> BinaryExpression:
        return TestCase.id == item_id

    @staticmethod
    def filter_criteria(
        pattern: Optional[str], suite_id: Optional[int], use_doc: bool, use_tags: bool
    ):
        filter_criteria = []
        if pattern:
            if use_tags:
                filter_criteria.append(TestCase.tags.ilike(glob_to_sql(pattern)))
            else:
                filter_criteria.append(TestCase.name.ilike(glob_to_sql(pattern)))
            if use_doc:
                filter_criteria = [
                    or_(filter_criteria[0], TestCase.doc.ilike(glob_to_sql(pattern)))
                ]
        if suite_id:
            filter_criteria.append(TestCase.suite_id == suite_id)
        return filter_criteria

    @staticmethod
    def from_row(row: Tuple[TestCase, Suite]) -> ModelTestCase:
        tc = row[0]
        return ModelTestCase(
            id=tc.id,
            name=tc.name,
            longname=f"{row[1].longname}.{tc.name}",
            line=tc.line,
            suite=row[1].to_nested(),
            doc=tc.doc,
            source=tc.source,
            template=tc.template,
            timeout=tc.timeout,
            keywords=KeywordRefList.parse_raw(tc.keywords),
            tags=TagList.parse_raw(tc.tags),
        )

    def get(self, item_id: int) -> Optional[ModelTestCase]:
        row = self._get_items.filter(self._id_filter(item_id)).first()
        if row:
            return self.from_row(row)

    def get_all(
        self,
        *,
        pattern: Optional[str] = None,
        suite_id: Optional[int] = None,
        use_doc: bool = True,
        use_tags: bool = False,
        skip: int = 0,
        limit: int = 100,
        ordering: List[OrderingItem] = None,
    ) -> List[ModelTestCase]:
        return [
            self.from_row(row)
            for row in (
                self._get_items.filter(
                    *self.filter_criteria(pattern, suite_id, use_doc, use_tags)
                )
                .order_by(*TestCase.ordering_criteria(ordering))
                .offset(skip)
                .limit(limit)
                .all()
            )
        ]

    def update(self, item: TestCase, update_data: dict) -> TestCase:
        if "keywords" in update_data:
            update_data["keywords"] = update_data["keywords"].json()
        if "tags" in update_data:
            update_data["tags"] = update_data["tags"].json()
        return super().update(item, update_data)



from abc import ABC
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.elements import BinaryExpression
from typing import Generic, Optional, TypeVar

from rfhub2.db.base import Base

T = TypeVar("T", bound=Base)


class BaseRepository(Generic[T], ABC):
    def __init__(self, db_session: Session):
        self.session = db_session

    @property
    def _items(self) -> Query:  # pragma: no cover
        raise NotImplementedError

    def add(self, item: T) -> T:
        self.session.add(item)
        self.session.commit()
        self.session.refresh(item)
        return item

    def update(self, item: T, update_data: dict):
        item_data = jsonable_encoder(item)
        for field in item_data:
            if field in update_data:
                setattr(item, field, update_data[field])
        return self.add(item)


class IdEntityRepository(BaseRepository, ABC):
    def _id_filter(self, item_id: int) -> BinaryExpression:  # pragma: no cover
        raise NotImplementedError

    def delete(self, item_id: int) -> int:
        row_count = self._items.filter(self._id_filter(item_id)).delete()
        self.session.commit()
        return row_count

    def get(self, item_id: int) -> Optional[T]:
        return self._items.get(item_id)



import uvicorn

from rfhub2 import config

# for now we import here this instance of app to have db initialized and populated for development purposes
# later we should just create app instance here
from rfhub2.main import app


def main():
    uvicorn.run(
        app,
        host=config.APP_INTERFACE,
        port=config.APP_PORT,
        log_level=config.APP_LOG_LEVEL,
    )


if __name__ == "__main__":
    main()



from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request

from rfhub2.db.session import Session


class DbSessionMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint):
        request.state.db = Session()
        response = await call_next(request)
        request.state.db.close()
        return response



from starlette.requests import Request

from rfhub2.db.repository.keyword_repository import KeywordRepository
from rfhub2.db.repository.collection_repository import CollectionRepository
from rfhub2.db.repository.keyword_statistics_repository import (
    KeywordStatisticsRepository,
)


def get_collection_repository(request: Request) -> CollectionRepository:
    return CollectionRepository(request.state.db)


def get_keyword_repository(request: Request) -> KeywordRepository:
    return KeywordRepository(request.state.db)


def get_keyword_statistics_repository(request: Request) -> KeywordStatisticsRepository:
    return KeywordStatisticsRepository(request.state.db)


def db_healthcheck(request: Request) -> bool:
    try:
        result = request.state.db.execute("select 1")
        return next(result) == (1,)
    except Exception as e:
        print(e)
        return False



from fastapi import Depends, HTTPException
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from starlette.status import HTTP_401_UNAUTHORIZED

from rfhub2 import config

security = HTTPBasic()


def authenticated_user(
    user: HTTPBasicCredentials = Depends(security)
) -> HTTPBasicCredentials:
    if (
        user.username == config.BASIC_AUTH_USER
        and user.password == config.BASIC_AUTH_PASSWORD
    ):
        return user
    else:
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail="Unauthorized to perform this action",
            headers={"WWW-Authenticate": "Basic"},
        )


def is_authenticated(user: HTTPBasicCredentials = Depends(security)) -> bool:
    return authenticated_user(user) is not None



from datetime import datetime
from enum import Enum
from pydantic import BaseModel, validator
from typing import List, Optional


class VersionInfo(BaseModel):
    title: str
    version: str


class Healthcheck(BaseModel):
    db: str


class TagList(BaseModel):
    __root__: List[str]

    @staticmethod
    def of(items: List[str]) -> "TagList":
        return TagList(__root__=items)


class NestedCollection(BaseModel):
    id: int
    name: str

    class Config:
        orm_mode = True


class KeywordUpdate(BaseModel):
    name: str
    doc: Optional[str]
    args: Optional[str]
    tags: Optional[List[str]]

    class Config:
        orm_mode = True


class KeywordCreate(KeywordUpdate):
    collection_id: int


class CollectionUpdate(BaseModel):
    name: str
    type: Optional[str]
    version: Optional[str]
    scope: Optional[str]
    named_args: Optional[str]
    path: Optional[str]
    doc: Optional[str]
    doc_format: Optional[str]


class NestedKeyword(KeywordUpdate):
    id: int
    synopsis: Optional[str]
    html_doc: Optional[str]
    arg_string: Optional[str]
    tags: List[str]


class Collection(NestedCollection, CollectionUpdate):
    keywords: List[NestedKeyword]
    synopsis: Optional[str]
    html_doc: Optional[str]


class CollectionWithStats(Collection):
    times_used: Optional[int]


class Keyword(NestedKeyword):
    collection: NestedCollection


class KeywordWithStats(Keyword):
    times_used: Optional[int]
    avg_elapsed: Optional[float]


class KeywordStatistics(BaseModel):
    collection: str
    keyword: str
    execution_time: datetime
    times_used: int
    total_elapsed: int
    min_elapsed: int
    max_elapsed: int

    class Config:
        orm_mode = True


class KeywordStatisticsList(BaseModel):
    __root__: List[KeywordStatistics]

    @staticmethod
    def of(items: List[KeywordStatistics]) -> "KeywordStatisticsList":
        return KeywordStatisticsList(__root__=items)


class StatisticsDeleted(BaseModel):
    deleted: int


class StatisticsInserted(BaseModel):
    inserted: int


class KeywordType(str, Enum):
    SETUP = "SETUP"
    NORMAL = "NORMAL"
    TEARDOWN = "TEARDOWN"


class KeywordRef(BaseModel):
    name: str
    args: List[str]
    kw_type: KeywordType


class KeywordRefList(BaseModel):
    __root__: List[KeywordRef]

    @staticmethod
    def of(items: List[KeywordRef]) -> "KeywordRefList":
        return KeywordRefList(__root__=items)


class MetadataItem(BaseModel):
    key: str
    value: str


class SuiteMetadata(BaseModel):
    __root__: List[MetadataItem]

    @staticmethod
    def of(items: List[MetadataItem]) -> "SuiteMetadata":
        return SuiteMetadata(__root__=items)


class Suite(BaseModel):
    id: int
    name: str
    longname: str
    doc: Optional[str]
    source: Optional[str]
    is_root: bool
    parent_id: Optional[int]
    test_count: int
    keywords: KeywordRefList
    metadata: SuiteMetadata
    rpa: bool = False


class SuiteHierarchy(BaseModel):
    name: str
    doc: Optional[str]
    source: Optional[str]
    keywords: KeywordRefList
    suites: List["SuiteHierarchy"]
    metadata: SuiteMetadata
    rpa: bool = False

    @validator("name")
    def name_cannot_have_dots(cls, v):
        if "." in v:
            raise ValueError("cannot have dots")
        return v


SuiteHierarchy.update_forward_refs()


class SuiteHierarchyWithId(BaseModel):
    id: int
    name: str
    longname: str
    doc: Optional[str]
    source: Optional[str]
    is_root: bool
    keywords: KeywordRefList
    suites: List["SuiteHierarchyWithId"]
    metadata: SuiteMetadata
    rpa: bool = False

    def with_suites(
        self, suites: List["SuiteHierarchyWithId"]
    ) -> "SuiteHierarchyWithId":
        self.suites = suites
        return self


SuiteHierarchyWithId.update_forward_refs()


class TestCaseCreate(BaseModel):
    name: str
    line: int
    suite_id: int
    doc: Optional[str]
    source: Optional[str]
    template: Optional[str]
    timeout: Optional[str]
    keywords: KeywordRefList
    tags: TagList


class TestCaseUpdate(BaseModel):
    name: Optional[str]
    line: Optional[int]
    suite_id: Optional[int]
    doc: Optional[str]
    source: Optional[str]
    template: Optional[str]
    timeout: Optional[str]
    keywords: Optional[KeywordRefList]
    tags: Optional[TagList]


class NestedSuite(BaseModel):
    id: int
    name: str
    longname: str


class TestCase(BaseModel):
    id: int
    name: str
    longname: str
    line: int
    suite: NestedSuite
    doc: Optional[str]
    source: Optional[str]
    template: Optional[str]
    timeout: Optional[str]
    keywords: KeywordRefList
    tags: TagList



from dataclasses import dataclass
from importlib.util import find_spec
from pathlib import Path
import re
from robot.errors import DataError
from robot.libdocpkg import LibraryDocumentation
from robot.libdocpkg.model import LibraryDoc
import robot.libraries
from robot.model import Tags
from typing import List, Optional, Set, Tuple, Union

from rfhub2.model import CollectionUpdate, KeywordUpdate

RESOURCE_PATTERNS = {".robot", ".txt", ".tsv", ".resource"}
ALL_PATTERNS = RESOURCE_PATTERNS | {".xml", ".py"}

INIT_FILES = {"__init__.txt", "__init__.robot", "__init__.html", "__init__.tsv"}

EXCLUDED_LIBRARIES = {
    "remote.py",
    "reserved.py",
    "dialogs.py",
    "dialogs_jy.py",
    "dialogs_py.py",
    "dialogs_ipy.py",
    "setup.py",
}


@dataclass
class CollectionUpdateWithKeywords:
    collection: CollectionUpdate
    keywords: List[KeywordUpdate]


class KeywordsExtractor:
    def __init__(
        self, paths: Tuple[Union[Path, str], ...], no_installed_keywords: bool
    ) -> None:
        self.paths = paths
        self.no_installed_keywords = no_installed_keywords

    def get_libraries_paths(self) -> Set[Path]:
        """
        Traverses all given paths and returns set with paths
        pointing to libraries to import to app.
        :return: Set of Paths object pointing to libraries to import
        """
        libraries_paths = set()
        for path in self.paths:
            path = self.get_library_path_from_name(path)
            if path:
                libraries_paths.update(self._traverse_paths(Path(path)))
        if not self.no_installed_keywords:
            libdir = Path(robot.libraries.__file__).parent
            libraries_paths.update(self._traverse_paths(Path(libdir)))
        return libraries_paths

    def get_library_path_from_name(self, path: Union[Path, str]) -> Optional[str]:
        """
        Helper function to recognize if given value is path or name.
        Name needs to be converted to full path and passed to Libdoc.
        """
        if Path(path).exists():
            return str(path)
        elif isinstance(path, str):
            try:
                return find_spec(path).submodule_search_locations[0]
            except AttributeError as e:
                print(
                    f"Collection {path} was neither valid path nor valid module name."
                )
                return None

    def _traverse_paths(self, path: Path) -> Set[Path]:
        """
        Traverses through paths and adds libraries to rfhub.
        Helper function for get_library_paths.
        """
        valid_lib_paths = set()
        if self._is_library_with_init(path):
            valid_lib_paths.add(path)
        else:
            for item in path.iterdir():
                if item.is_dir():
                    if self._is_library_with_init(item):
                        valid_lib_paths.add(item)
                        if self._robot_files_candidates(item):
                            valid_lib_paths.update(self._get_valid_robot_files(path))
                    else:
                        valid_lib_paths.update(self._traverse_paths(item))
                elif (
                    item.is_file()
                    and self._is_robot_keyword_file(item)
                    and not self._should_ignore(item)
                ):
                    valid_lib_paths.add(item)
        return valid_lib_paths

    def create_collections(
        self, paths: Set[Path]
    ) -> List[CollectionUpdateWithKeywords]:
        """
        Creates list of Collection objects from set of provided paths.
        :param paths: set of paths
        :return: list of Collection objects
        """
        collections = []
        for path in paths:
            try:
                collection_with_keywords = self.create_collection(path)
                collections.append(collection_with_keywords)
            except (DataError, SystemExit) as ex:
                print(
                    f"Failed to create collection from path {path}\n"
                    f"{type(ex).__name__}, {ex.args}"
                )
        return sorted(collections, key=lambda i: i.collection.name)

    def create_collection(self, path: Path) -> CollectionUpdateWithKeywords:
        """
        Creates CollectionUpdateWithKeywords object from provided path.
        :param path: Path
        :return: CollectionUpdateWithKeywords object
        """
        libdoc = LibraryDocumentation(str(path))
        return CollectionUpdateWithKeywords(
            self._serialise_libdoc(libdoc, str(path)), self._serialise_keywords(libdoc)
        )

    def _serialise_libdoc(self, libdoc: LibraryDoc, path: str) -> CollectionUpdate:
        """
        Serialises LibraryDoc object to CollectionUpdate object.
        :param libdoc: LibraryDoc input object
        :param path: library path
        :return: CollectionUpdate object
        """
        return CollectionUpdate(
            name=libdoc.name,
            type=libdoc.type,
            version=libdoc.version,
            scope=libdoc.scope,
            # named_args=libdoc.named_args, # we have not used this one, yet
            path=path,
            doc=libdoc.doc + self._extract_doc_from_libdoc_inits(libdoc.inits),
            doc_format=libdoc._setter__doc_format,
        )

    def _serialise_keywords(self, libdoc: LibraryDoc) -> List[KeywordUpdate]:
        """
        Serialises keywords to KeywordUpdate object.
        :param :LibraryDoc input object
        :return: KeywordUpdate object
        """
        return [
            KeywordUpdate(
                name=keyword.name,
                args=self._serialise_args(keyword.args),
                tags=self._serialise_tags(keyword.tags),
                doc=keyword.doc,
            )
            for keyword in libdoc.keywords
        ]

    def _serialise_args(self, args: List[str]) -> str:
        return (
            str([str(item).replace("'", "").replace('"', "") for item in args]).replace(
                "'", '"'
            )
            if args
            else ""
        )

    def _serialise_tags(self, tags: Tags) -> List[str]:
        return list(tags._tags)

    def _extract_doc_from_libdoc_inits(self, inits: List) -> str:
        return "\n" + "\n" + "\n".join([d.doc for d in inits]) if len(inits) > 0 else ""

    def _robot_files_candidates(self, path: Path) -> bool:
        return (
            len(
                [file for file in path.glob("**/*") if file.suffix in RESOURCE_PATTERNS]
            )
            > 0
        )

    def _get_valid_robot_files(self, path: Path) -> Set[Path]:
        return {
            file
            for file in path.glob("**/*")
            if (self._is_libdoc_file(file) or self._is_resource_file(file))
        }

    @staticmethod
    def _is_library_with_init(path: Path) -> bool:
        return (path / "__init__.py").is_file() and len(
            LibraryDocumentation(str(path)).keywords
        ) > 0

    def _is_robot_keyword_file(self, file: Path) -> bool:
        return (
            self._is_library_file(file)
            or self._is_libdoc_file(file)
            or self._is_resource_file(file)
        )

    @staticmethod
    def _is_library_file(file: Path) -> bool:
        return file.suffix == ".py" and file.name != "__init__.py"

    @staticmethod
    def _is_libdoc_file(file: Path) -> bool:
        """Return true if an xml file looks like a libdoc file"""
        # inefficient since we end up reading the file twice,
        # but it's fast enough for our purposes, and prevents
        # us from doing a full parse of files that are obviously
        # not libdoc files
        if file.suffix == ".xml":
            with open(file, "r", encoding="utf-8", errors="ignore") as f:
                # read the first few lines; if we don't see
                # what looks like libdoc data, return false
                data = f.read(200)
                return "<keywordspec " in data.lower()
        return False

    @staticmethod
    def _should_ignore(file: Path) -> bool:
        """Return True if a given library name should be ignored
        This is necessary because not all files we find in the library
        folder are libraries.
        """
        filename = file.name.lower()
        return (
            filename.startswith("deprecated")
            or filename.startswith("_")
            or filename in EXCLUDED_LIBRARIES
        )

    @staticmethod
    def _is_resource_file(file: Path) -> bool:
        """Returns true if the file has a keyword table but not a testcase table."""
        # inefficient since we end up reading the file twice,
        # but it's fast enough for our purposes, and prevents
        # us from doing a full parse of files that are obviously
        # not robot files

        if file.name not in INIT_FILES and file.suffix in RESOURCE_PATTERNS:
            with open(file, "r", encoding="utf-8", errors="ignore") as f:
                data = f.read()
                return not KeywordsExtractor._has_test_case_table(
                    data
                ) and KeywordsExtractor._has_keyword_table(data)
        return False

    @staticmethod
    def _has_keyword_table(data: str) -> bool:
        """Returns true if file has keyword or user keyword table"""
        return (
            re.search(
                r"^\*+\s*((?:User )?Keywords?)", data, re.MULTILINE | re.IGNORECASE
            )
            is not None
        )

    @staticmethod
    def _has_test_case_table(data: str) -> bool:
        """Returns true if file has keyword or user keyword table"""
        return (
            re.search(r"^\*+\s*(Test Cases?)", data, re.MULTILINE | re.IGNORECASE)
            is not None
        )



"""Add suite model

Revision ID: d3ab100869e0
Revises: abee96e78502
Create Date: 2020-11-28 22:34:10.042146

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "d3ab100869e0"
down_revision = "abee96e78502"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "suite",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("longname", sa.Text(), nullable=False),
        sa.Column("doc", sa.Text(), nullable=True),
        sa.Column("source", sa.Text(), nullable=True),
        sa.Column("keywords", sa.Text(), nullable=False),
        sa.Column("metadata_items", sa.Text(), nullable=False),
        sa.Column("is_root", sa.Boolean(), nullable=False),
        sa.Column("rpa", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_suite_name"), "suite", ["name"], unique=False)
    op.create_index(op.f("ix_suite_longname"), "suite", ["longname"], unique=True)
    op.create_table(
        "suiterel",
        sa.Column("parent_id", sa.Integer(), nullable=False),
        sa.Column("child_id", sa.Integer(), nullable=False),
        sa.Column("degree", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("parent_id", "child_id"),
        sa.ForeignKeyConstraint(["parent_id"], ["suite.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(["child_id"], ["suite.id"], ondelete="CASCADE"),
    )
    op.create_table(
        "testcase",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("suite_id", sa.Integer(), nullable=False),
        sa.Column("line", sa.Integer(), nullable=False),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("doc", sa.Text(), nullable=True),
        sa.Column("source", sa.Text(), nullable=True),
        sa.Column("template", sa.Text(), nullable=True),
        sa.Column("timeout", sa.Text(), nullable=True),
        sa.Column("tags", sa.Text(), nullable=False),
        sa.Column("keywords", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.ForeignKeyConstraint(["suite_id"], ["suite.id"], ondelete="CASCADE"),
    )
    op.create_index(op.f("ix_testcase_name"), "testcase", ["name"], unique=False)
    op.create_index(
        op.f("ix_testcase_suite_id_name"), "testcase", ["suite_id", "name"], unique=True
    )
    # ### end Alembic commands ###


def downgrade():
    op.drop_index(op.f("ix_testcase_suite_id_name"), table_name="testcase")
    op.drop_index(op.f("ix_testcase_name"), table_name="testcase")
    op.drop_table("testcase")
    op.drop_table("suiterel")
    op.drop_index(op.f("ix_suite_longname"), table_name="suite")
    op.drop_index(op.f("ix_suite_name"), table_name="suite")
    op.drop_table("suite")
    # ### end Alembic commands ###



from sqlalchemy import Column, Integer, Sequence, Text
from sqlalchemy.orm import relationship
from typing import List

from rfhub2.db.model.base_class import Base
from rfhub2.db.model.mixins import DocMixin
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2 import model


class Collection(Base, DocMixin):
    id = Column(Integer, Sequence("collection_id_seq"), primary_key=True)
    name = Column(Text, index=True)
    type = Column(Text)
    version = Column(Text)
    scope = Column(Text)
    named_args = Column(Text)
    path = Column(Text)
    doc = Column(Text)
    doc_format = Column(Text)
    keywords = relationship(
        "Keyword",
        backref="collection",
        cascade="all, delete-orphan",
        passive_deletes=True,
        order_by="Keyword.name",
    )

    def __str__(self):  # pragma: no cover
        return (
            f"Collection({self.id},{self.name},{self.type},{self.version},"
            + f"{self.scope},{self.named_args},{self.path},{self.doc},{self.doc_format})"
        )

    def __repr__(self):  # pragma: no cover
        return str(self)

    @classmethod
    def default_ordering(cls) -> List[OrderingItem]:
        return [OrderingItem("name")]

    @staticmethod
    def create(data: model.CollectionUpdate) -> "Collection":
        return Collection(
            name=data.name,
            type=data.type,
            version=data.version,
            scope=data.scope,
            named_args=data.named_args,
            path=data.path,
            doc=data.doc,
            doc_format=data.doc_format,
        )

    def to_model(self) -> model.Collection:
        keywords = [kw.to_nested_model() for kw in self.keywords]
        return model.Collection(
            id=self.id,
            name=self.name,
            type=self.type,
            version=self.version,
            scope=self.scope,
            named_args=self.named_args,
            path=self.path,
            doc=self.doc,
            doc_format=self.doc_format,
            html_doc=self.html_doc,
            synopsis=self.synopsis,
            keywords=keywords,
        )

    def to_nested_model(self) -> model.NestedCollection:
        return model.NestedCollection(id=self.id, name=self.name)



import json
from sqlalchemy import Column, ForeignKey, Integer, Sequence, Text
from typing import List, Optional

from rfhub2.db.model.base_class import Base
from rfhub2.db.model.mixins import KeywordMixin
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2 import model


class Keyword(Base, KeywordMixin):
    id = Column(Integer, Sequence("keyword_id_seq"), primary_key=True)
    name = Column(Text, index=True)
    doc = Column(Text)
    args = Column(Text)
    tags = Column(Text)
    collection_id = Column(
        Integer, ForeignKey("collection.id", ondelete="CASCADE"), nullable=False
    )

    def __str__(self):  # pragma: no cover
        return f"Keyword({self.id},{self.name},{self.doc},{self.args},{self.tags},{self.collection_id})"

    def __repr__(self):  # pragma: no cover
        return str(self)

    @classmethod
    def default_ordering(cls) -> List[OrderingItem]:
        return [OrderingItem("name")]

    @staticmethod
    def create(data: model.KeywordCreate) -> "Keyword":
        return Keyword(
            name=data.name,
            doc=data.doc,
            args=data.args,
            tags=Keyword.to_json_list(data.tags),
            collection_id=data.collection_id,
        )

    def to_model(self) -> model.Keyword:
        collection = self.collection
        return model.Keyword(
            id=self.id,
            name=self.name,
            doc=self.doc,
            args=self.args,
            tags=Keyword.from_json_list(self.tags),
            arg_string=self.arg_string,
            html_doc=self.html_doc,
            synopsis=self.synopsis,
            collection=collection.to_nested_model(),
        )

    def to_nested_model(self) -> model.NestedKeyword:
        return model.NestedKeyword(
            id=self.id,
            name=self.name,
            doc=self.doc,
            args=self.args,
            tags=Keyword.from_json_list(self.tags),
            arg_string=self.arg_string,
            html_doc=self.html_doc,
            synopsis=self.synopsis,
        )

    @staticmethod
    def from_json_list(json_list: Optional[str]) -> List[str]:
        return json.loads(json_list) if json_list else []

    @staticmethod
    def to_json_list(items: Optional[List[str]]) -> str:
        item_list = items or []
        return json.dumps(item_list)



from fastapi import APIRouter, Depends, HTTPException
from starlette.responses import Response
from typing import List, Optional

from rfhub2.api.utils.auth import is_authenticated
from rfhub2.api.utils.db import get_collection_repository
from rfhub2.api.utils.http import or_404
from rfhub2.api.utils.order import get_ordering
from rfhub2.db.base import Collection as DBCollection
from rfhub2.db.repository.collection_repository import CollectionRepository
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2.model import Collection, CollectionUpdate, CollectionWithStats

router = APIRouter()


@router.get("/", response_model=List[Collection])
def get_collections(
    repository: CollectionRepository = Depends(get_collection_repository),
    skip: int = 0,
    limit: int = 100,
    pattern: str = None,
    libtype: str = None,
    ordering: List[OrderingItem] = Depends(get_ordering),
):
    return repository.get_all(
        skip=skip, limit=limit, pattern=pattern, libtype=libtype, ordering=ordering
    )


@router.get("/stats/", response_model=List[CollectionWithStats])
def get_collections_with_stats(
    repository: CollectionRepository = Depends(get_collection_repository),
    skip: int = 0,
    limit: int = 100,
    pattern: str = None,
    libtype: str = None,
    ordering: List[OrderingItem] = Depends(get_ordering),
):
    return repository.get_all_with_stats(
        skip=skip, limit=limit, pattern=pattern, libtype=libtype, ordering=ordering
    )


@router.get("/stats/{id}/", response_model=CollectionWithStats)
def get_collection_with_stats(
    *, repository: CollectionRepository = Depends(get_collection_repository), id: int
):
    collection: Optional[CollectionWithStats] = repository.get_with_stats(id)
    return or_404(collection)


@router.get("/{id}/", response_model=Collection)
def get_collection(
    *, repository: CollectionRepository = Depends(get_collection_repository), id: int
):
    collection: Optional[DBCollection] = repository.get(id)
    return or_404(collection).to_model()


@router.post("/", response_model=Collection, status_code=201)
def create_collection(
    *,
    _: bool = Depends(is_authenticated),
    repository: CollectionRepository = Depends(get_collection_repository),
    collection: CollectionUpdate,
):
    db_collection: DBCollection = repository.add(DBCollection.create(collection))
    return db_collection.to_model()


@router.put("/{id}/", response_model=Collection)
def update_collection(
    *,
    _: bool = Depends(is_authenticated),
    repository: CollectionRepository = Depends(get_collection_repository),
    id: int,
    collection_update: CollectionUpdate,
):
    db_collection: DBCollection = or_404(repository.get(id))
    updated: DBCollection = repository.update(
        db_collection, collection_update.dict(exclude_unset=True)
    )
    return updated.to_model()


@router.delete("/{id}/")
def delete_collection(
    *,
    _: bool = Depends(is_authenticated),
    repository: CollectionRepository = Depends(get_collection_repository),
    id: int,
):
    deleted: int = repository.delete(id)
    if deleted:
        return Response(status_code=204)
    else:
        raise HTTPException(status_code=404)



import json

try:
    from robot.libdocpkg.htmlutils import DocToHtml
except ImportError:
    from robot.libdocpkg.htmlwriter import DocToHtml


class DocMixin:
    @property
    def synopsis(self) -> str:
        return self.doc.splitlines()[0] if self.doc else ""

    @property
    def html_doc(self) -> str:
        return DocToHtml("ROBOT")(self.doc) if self.doc else ""


class KeywordMixin(DocMixin):
    @property
    def arg_string(self) -> str:
        """
        Old implementation saves args list as JSON in text field, this is more readable representation for UI
        """
        return ", ".join(json.loads(self.args)) if self.args else ""



from sqlalchemy import Column, DateTime, Integer, PrimaryKeyConstraint, Text
from typing import List

from rfhub2.db.model.base_class import Base
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2 import model


class KeywordStatistics(Base):
    collection = Column(Text)
    keyword = Column(Text)
    execution_time = Column(DateTime(timezone=True))
    times_used = Column(Integer)
    total_elapsed = Column(Integer)
    min_elapsed = Column(Integer)
    max_elapsed = Column(Integer)

    __table_args__ = (PrimaryKeyConstraint(collection, keyword, execution_time),)

    def __str__(self):  # pragma: no cover
        return (
            f"KeywordStatistics({self.collection},{self.keyword},{self.execution_time},"
            + f"{self.times_used},{self.total_elapsed},{self.min_elapsed},{self.max_elapsed})"
        )

    def __repr__(self):  # pragma: no cover
        return str(self)

    @classmethod
    def default_ordering(cls) -> List[OrderingItem]:
        return [
            OrderingItem("collection"),
            OrderingItem("keyword"),
            OrderingItem("execution_time", False),
        ]

    @staticmethod
    def create(data: model.KeywordStatistics) -> "KeywordStatistics":
        return KeywordStatistics(
            collection=data.collection,
            keyword=data.keyword,
            execution_time=data.execution_time,
            times_used=data.times_used,
            total_elapsed=data.total_elapsed,
            min_elapsed=data.min_elapsed,
            max_elapsed=data.max_elapsed,
        )

    def to_model(self) -> model.KeywordStatistics:
        return model.KeywordStatistics(
            collection=self.collection,
            keyword=self.keyword,
            execution_time=self.execution_time,
            times_used=self.times_used,
            total_elapsed=self.total_elapsed,
            min_elapsed=self.min_elapsed,
            max_elapsed=self.max_elapsed,
        )



from fastapi import APIRouter

from rfhub2.config import APP_TITLE
from rfhub2.model import VersionInfo
from rfhub2.version import version

router = APIRouter()


@router.get("/", response_model=VersionInfo)
def healthcheck():
    return VersionInfo(**{"title": APP_TITLE, "version": version})



from sqlalchemy import Column, ForeignKey, Integer, PrimaryKeyConstraint

from rfhub2.db.model.base_class import Base


class SuiteRel(Base):
    parent_id = Column(Integer, ForeignKey("suite.id", ondelete="CASCADE"))
    child_id = Column(Integer, ForeignKey("suite.id", ondelete="CASCADE"))
    degree = Column(Integer)

    __table_args__ = (PrimaryKeyConstraint(parent_id, child_id),)

    def __str__(self):  # pragma: no cover
        return f"SuiteRel({self.parent_id},{self.child_id},{self.degree})"

    def __repr__(self):  # pragma: no cover
        return str(self)



from typing import List, Optional, Tuple

from sqlalchemy import and_, or_, func
from sqlalchemy.orm import selectinload
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.elements import BinaryExpression

from rfhub2.db.base import Collection, Keyword, KeywordStatistics
from rfhub2.model import KeywordWithStats, Keyword as ModelKeyword
from rfhub2.db.repository.base_repository import IdEntityRepository
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2.db.repository.query_utils import glob_to_sql


class KeywordRepository(IdEntityRepository):
    @property
    def _items(self) -> Query:
        return self.session.query(Keyword).options(selectinload(Keyword.collection))

    @property
    def _items_with_stats(self) -> Query:
        keyword_statistics = (
            self.session.query(
                func.sum(KeywordStatistics.times_used).label("times_used"),
                (
                    func.sum(KeywordStatistics.total_elapsed)
                    / func.sum(KeywordStatistics.times_used)
                ).label("avg_elapsed"),
                KeywordStatistics.collection,
                KeywordStatistics.keyword,
            )
            .group_by(KeywordStatistics.collection, KeywordStatistics.keyword)
            .subquery()
        )
        return (
            self.session.query(
                Keyword,
                keyword_statistics.c.times_used,
                keyword_statistics.c.avg_elapsed,
            )
            .options(selectinload(Keyword.collection))
            .join(Keyword.collection)
            .outerjoin(
                keyword_statistics,
                and_(
                    Collection.name == keyword_statistics.c.collection,
                    Keyword.name == keyword_statistics.c.keyword,
                ),
            )
        )

    def _id_filter(self, item_id: int) -> BinaryExpression:
        return Keyword.id == item_id

    @staticmethod
    def filter_criteria(
        pattern: Optional[str],
        collection_name: Optional[str],
        collection_id: Optional[int],
        use_doc: bool,
        use_tags: bool,
    ):
        filter_criteria = []
        if pattern:
            if use_tags:
                filter_criteria.append(Keyword.tags.ilike(glob_to_sql(pattern)))
            else:
                filter_criteria.append(Keyword.name.ilike(glob_to_sql(pattern)))
            if use_doc:
                filter_criteria = [
                    or_(filter_criteria[0], Keyword.doc.ilike(glob_to_sql(pattern)))
                ]
        if collection_name:
            filter_criteria.append(Collection.name.ilike(glob_to_sql(collection_name)))
        if collection_id:
            filter_criteria.append(Keyword.collection_id == collection_id)
        return filter_criteria

    @staticmethod
    def from_stats_row(row: Tuple[Keyword, int, float]) -> KeywordWithStats:
        keyword = row[0]
        return KeywordWithStats(
            id=keyword.id,
            name=keyword.name,
            doc=keyword.doc,
            args=keyword.args,
            tags=Keyword.from_json_list(keyword.tags),
            arg_string=keyword.arg_string,
            html_doc=keyword.html_doc,
            synopsis=keyword.synopsis,
            collection=keyword.collection.to_nested_model(),
            times_used=row[1],
            avg_elapsed=row[2],
        )

    def get_all(
        self,
        *,
        pattern: Optional[str] = None,
        collection_name: Optional[str] = None,
        collection_id: Optional[int] = None,
        use_doc: bool = True,
        use_tags: bool = False,
        skip: int = 0,
        limit: int = 100,
        ordering: List[OrderingItem] = None,
    ) -> List[ModelKeyword]:
        return [
            keyword.to_model()
            for keyword in (
                self.session.query(Keyword)
                .join(Keyword.collection)
                .filter(
                    *self.filter_criteria(
                        pattern, collection_name, collection_id, use_doc, use_tags
                    )
                )
                .order_by(*Keyword.ordering_criteria(ordering))
                .offset(skip)
                .limit(limit)
                .all()
            )
        ]

    def get_all_with_stats(
        self,
        *,
        pattern: Optional[str] = None,
        collection_name: Optional[str] = None,
        collection_id: Optional[int] = None,
        use_doc: bool = True,
        use_tags: bool = False,
        skip: int = 0,
        limit: int = 100,
        ordering: List[OrderingItem] = None,
    ) -> List[KeywordWithStats]:
        return [
            self.from_stats_row(row)
            for row in (
                self._items_with_stats.filter(
                    *self.filter_criteria(
                        pattern, collection_name, collection_id, use_doc, use_tags
                    )
                )
                .order_by(*Keyword.ordering_criteria(ordering))
                .offset(skip)
                .limit(limit)
                .all()
            )
        ]

    def get_with_stats(self, item_id: int) -> Optional[KeywordWithStats]:
        result = self._items_with_stats.filter(self._id_filter(item_id)).first()
        if result:
            return self.from_stats_row(result)

    def update(self, item: Keyword, update_data: dict) -> Keyword:
        if "tags" in update_data:
            update_data["tags"] = Keyword.from_json_list(update_data["tags"])
        return super().update(item, update_data)



def glob_to_sql(string: str) -> str:
    """Convert glob-like wildcards to SQL wildcards

    * becomes %
    ? becomes _
    % becomes \%
    \\ remains \\
    \* remains \*
    \? remains \?

    This also adds a leading and trailing %, unless the pattern begins with
    ^ or ends with $
    """

    # What's with the chr(1) and chr(2) nonsense? It's a trick to
    # hide \* and \? from the * and ? substitutions. This trick
    # depends on the substitutions being done in order.  chr(1)
    # and chr(2) were picked because I know those characters
    # almost certainly won't be in the input string
    table = (
        (r"\\", chr(1)),
        (r"\*", chr(2)),
        (r"\?", chr(3)),
        (r"%", r"\%"),
        (r"?", "_"),
        (r"*", "%"),
        (chr(1), r"\\"),
        (chr(2), r"\*"),
        (chr(3), r"\?"),
    )

    for (a, b) in table:
        string = string.replace(a, b)

    string = string[1:] if string.startswith("^") else "%" + string
    string = string[:-1] if string.endswith("$") else string + "%"

    return string



"""create tags column for keyword table

Revision ID: abee96e78502
Revises: c54a916ec6c8
Create Date: 2020-04-04 12:50:17.064642

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "abee96e78502"
down_revision = "c54a916ec6c8"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("keyword", sa.Column("tags", sa.Text, nullable=True))


def downgrade():
    op.drop_column("keyword", "tags")



from sqlalchemy import Column, ForeignKey, Index, Integer, Sequence, Text
from typing import List

from rfhub2.db.model.base_class import Base
from rfhub2.db.model.mixins import DocMixin
from rfhub2.db.repository.ordering import OrderingItem
from rfhub2.model import TestCaseCreate


class TestCase(Base, DocMixin):
    id = Column(Integer, Sequence("test_case_id_seq"), primary_key=True)
    suite_id = Column(
        Integer, ForeignKey("suite.id", ondelete="CASCADE"), nullable=False
    )
    line = Column(Integer)
    name = Column(Text, index=True)
    doc = Column(Text, nullable=True)
    source = Column(Text, nullable=True)
    template = Column(Text, nullable=True)
    timeout = Column(Text, nullable=True)
    tags = Column(Text)
    keywords = Column(Text)

    __table_args__ = (Index("ix_testcase_suite_id_name", suite_id, name, unique=True),)

    def __str__(self):  # pragma: no cover
        return (
            f"TestCase({self.id},{self.suite_id},{self.name},{self.line},{self.doc},"
            + f"{self.source},{self.template},{self.timeout},{self.tags},{self.keywords})"
        )

    def __repr__(self):  # pragma: no cover
        return str(self)

    @classmethod
    def default_ordering(cls) -> List[OrderingItem]:
        return [OrderingItem("name")]

    @staticmethod
    def create(tc: TestCaseCreate) -> "TestCase":
        return TestCase(
            suite_id=tc.suite_id,
            line=tc.line,
            name=tc.name,
            doc=tc.doc,
            source=tc.source,
            template=tc.template,
            timeout=tc.timeout,
            keywords=tc.keywords.json(),
            tags=tc.tags.json(),
        )
